# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scheduled GP-Bandit for budget-aware optimization.
"""

from typing import Callable
import attrs
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import gp_bandit
from vizier._src.algorithms.designers import scheduled_designer
from vizier._src.algorithms.designers.gp import acquisitions


@attrs.define(kw_only=True)
class ScheduledGPBanditFactory:
  """Scheduled GP-Bandit factory."""

  _gp_bandit_factory: Callable[[vz.ProblemStatement], gp_bandit.VizierGPBandit]
  _init_ucb_coefficient: float
  _final_ucb_coefficient: float
  _decay_ucb_coefficient: float
  _expected_total_num_trials: int

  def __call__(
      self,
      problem: vz.ProblemStatement,
  ) -> scheduled_designer.ScheduledDesigner:
    """Creates a scheduled GP-Bandit designer."""

    def _gp_bandit_state_updater(designer, params):
      designer._scoring_function_factory = (  # pylint: disable=protected-access
          acquisitions.bayesian_scoring_function_factory(
              lambda _: acquisitions.UCB(params['ucb_coefficient'])
          )
      )

    ucb_coef_param = scheduled_designer.ExponentialScheduledParam(
        init_value=self._init_ucb_coefficient,
        final_value=self._final_ucb_coefficient,
        rate=self._decay_ucb_coefficient,
    )

    return scheduled_designer.ScheduledDesigner(
        problem,
        designer_factory=self._gp_bandit_factory,
        designer_state_updater=_gp_bandit_state_updater,
        scheduled_params={'ucb_coefficient': ucb_coef_param},
        expected_total_num_trials=self._expected_total_num_trials,
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Base class for acquisition optimizers."""

import abc
from typing import Any, Callable, Optional, Protocol, Sequence

import attr
from vizier import pyvizier as vz

Array = Any


def _is_positive(instance, attribute, value):
  del instance, attribute
  if value < 0:
    raise ValueError(f'value must be positive! given: {value}')


class BatchTrialScoreFunction(Protocol):
  """Protocol (https://peps.python.org/pep-0544/) for scoring trials."""

  # TODO: Decide what to do with NaNs.
  def __call__(self, trials: Sequence[vz.Trial]) -> dict[str, Array]:
    """Evaluates the trials.

    Args:
      trials: A sequence of N trials

    Returns:
      A dict of shape (N, 1) arrays.
    """


@attr.s(frozen=False, init=True, slots=True)
class BranchSelection:
  """Branch Selection.

  Decision of a branch selector that contains the flat subspace and number
  of suggestions to be generated in that subspace.

  Instead of creating N suggestions on a conditional space S, we create
  N_1, ..., N_k suggestions on flat spaces S_1, .., S_k such that
  N_1 + ... + N_k = N and S_1, ..., S_k are contained in S. Each BranchSelection
  object represents (N_i, S_i) tuple.

  Attributres:
    search_space: Search space that does not contain any conditional parameters.
    num_suggestions: Number of suggestions to be made in the search space.
  """
  search_space: vz.SearchSpace = attr.ib(
      validator=attr.validators.instance_of(vz.SearchSpace),
      on_setattr=attr.setters.validate)
  num_suggestions: int = attr.ib(
      validator=[attr.validators.instance_of(int), _is_positive],
      on_setattr=attr.setters.validate)


class BranchSelector(abc.ABC):

  @abc.abstractmethod
  def select_branches(self, num_suggestions: int) -> list[BranchSelection]:
    pass


class GradientFreeOptimizer(abc.ABC):
  """Optimizes a function on Vizier search space.

  Typically used for optimizing acquisition functions.
  """

  @abc.abstractmethod
  def optimize(
      self,
      score_fn: BatchTrialScoreFunction,
      problem: vz.ProblemStatement,
      *,
      count: int = 1,
      budget_factor: float = 1.0,
      seed_candidates: Sequence[vz.TrialSuggestion] = tuple()
  ) -> list[vz.Trial]:
    """Optimizes a function.

    Args:
      score_fn: Should return a dict whose keys contain the metric names of
        "problem.metric_information".
      problem:
      count: Optimizer tries to return this many trials.
      budget_factor:  For optimizers with a notion of a budget, use this much
        fraction of the standard budget for the call.
      seed_candidates: Seed suggestions to be used as initial batch for
        optimization.

    Returns:
      Trials, of length less than or equal to max_num_suggestions.
      Trials are COMPLETED with score_fn results.
    """
    pass


@attr.frozen
class BranchThenOptimizer(GradientFreeOptimizer):
  """Optimizes a function by first choosing a branch and then apply Optimizer.

  Attributes:
    _branch_selector: Selects all conditional parent values
    _optimizer_factory: Creates an optimizer in the flat (non-conditional) after
      branch selector fixing all conditional parent values.
    max_num_suggestions_per_branch: Limits the number of suggestions per branch.
      This is useful when acquisition function isn't well-suited for batch
      suggestions.
  """
  _branch_selector: BranchSelector
  _optimizer_factory: Callable[[], GradientFreeOptimizer]
  max_num_suggestions_per_branch: Optional[int] = None

  def _num_suggestions_for_branch(self, branch: BranchSelection) -> int:
    if self.max_num_suggestions_per_branch is None:
      return branch.num_suggestions
    else:
      return min(self.max_num_suggestions_per_branch, branch.num_suggestions)

  def optimize(
      self,
      score_fn: BatchTrialScoreFunction,
      problem: vz.ProblemStatement,
      *,
      count: int = 1,
      budget_factor: float = 1.0,
      seed_candidates: Sequence[vz.TrialSuggestion] = tuple(),
  ) -> list[vz.Trial]:
    # If there are conditional branches, use Vizier's default branch
    # selection mechanism.
    branches = self._branch_selector.select_branches(count)
    suggestions = []
    optimizer = self._optimizer_factory()
    for branch in branches:
      subproblem = attr.evolve(problem, search_space=branch.search_space)
      suggestions.extend(
          optimizer.optimize(
              score_fn,
              subproblem,
              count=self._num_suggestions_for_branch(branch),
              budget_factor=budget_factor * (branch.num_suggestions / count)))
    return suggestions



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for profiler."""

import equinox as eqx
import jax
import jax.numpy as jnp
from vizier.utils import profiler

from absl.testing import absltest


class MyModule(eqx.Module):

  @profiler.record_tracing
  def add_one(self, x: int) -> int:
    return x + 1


@profiler.record_tracing
def add_one(x):
  return x + 1


class PerformanceUtilsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    jax.clear_caches()

  def test_eqx_jit(self):
    """`record_tracing` is tested on a bound method that is `filter_jit`-ed."""

    @profiler.record_runtime(name='return_two', also_log=True)
    def return_two():
      one = eqx.filter_jit(MyModule().add_one)(jnp.array(0.0))
      two = eqx.filter_jit(MyModule().add_one)(one)
      return two

    with profiler.collect_events() as events:
      return_two()
      return_two()

    # tracing is logged once, timer is logged twice.
    self.assertLen(events, 3)

    self.assertEqual(events[0].etype, profiler.EventType.JIT_TRACING)
    self.assertEqual(events[1].etype, profiler.EventType.TIMER)
    self.assertEqual(events[2].etype, profiler.EventType.TIMER)

    # Tracing happens within the timed functions' scope.
    self.assertIn(events[1].scope, events[0].scope)

    latencies = profiler.get_latencies_dict(events)
    self.assertIn('return_two', latencies)
    self.assertLen(latencies['return_two'], 2)

  def test_regular_jit(self):
    """`record_tracing` is tested on an unbound method."""

    @profiler.record_runtime(name='return_two', also_log=True)
    def return_two():
      one = jax.jit(add_one)(jnp.array(0.0))
      two = jax.jit(add_one)(one)
      return two

    with profiler.collect_events() as events:
      return_two()

    # tracing is logged once, timer is logged once.
    self.assertLen(events, 2)
    self.assertEqual(events[0].etype, profiler.EventType.JIT_TRACING)
    self.assertEqual(events[1].etype, profiler.EventType.TIMER)

    # Tracing happens within the timed functions' scope.
    self.assertIn(events[1].scope, events[0].scope)


if __name__ == '__main__':
  jax.config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for jnp_converters."""

import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random as random_designer_lib
from vizier._src.algorithms.testing import test_runners
from vizier._src.jax import types
from vizier.pyvizier.converters import jnp_converters as jnpc
from vizier.pyvizier.converters import padding
from vizier.testing import test_studies

from absl.testing import absltest
from absl.testing import parameterized


Trial = vz.Trial


class PaddedTrialToArrayConverterTest(parameterized.TestCase):

  def test_padding(self):
    """Tests various padding schedules."""

    padding_schedule = padding.PaddingSchedule(
        num_trials=padding.PaddingType.POWERS_OF_2,
        num_features=padding.PaddingType.MULTIPLES_OF_10,
        num_metrics=padding.PaddingType.NONE,
    )
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric',
            goal=vz.ObjectiveMetricGoal.MAXIMIZE,
            min_value=-1.0,
            max_value=1.0,
        )
    )
    trials = test_runners.RandomMetricsRunner(
        problem,
        iters=1,
        batch_size=13,
        verbose=False,
        validate_parameters=False,
    ).run_designer(random_designer_lib.RandomDesigner(problem.search_space))

    converter = jnpc.TrialToModelInputConverter.from_problem(
        problem,
        padding_schedule=padding_schedule,
    )

    data = converter.to_xy(trials)
    self.assertSequenceEqual(
        data.features.continuous.padded_array.shape,
        (16, 10),
    )
    self.assertSequenceEqual(
        data.features.categorical.padded_array.shape,
        (16, 10),
    )

    labels = converter.to_labels(trials)
    self.assertSequenceEqual(labels.padded_array.shape, (16, 1))

    recovererd_trials = converter.to_trials(data)

    for i in range(13):
      self.assertSequenceAlmostEqual(
          recovererd_trials[i].parameters.as_dict().values(),
          trials[i].parameters.as_dict().values(),
          places=3,
          msg=f'Failed at {i}th trial',
      )
      problem.search_space.assert_contains(recovererd_trials[i].parameters)


class TrialToContinuousAndCategoricalConverterTest(parameterized.TestCase):
  """Test TrialToContinuousAndCategoricalConverter class."""

  def setUp(self):
    super().setUp()
    self._study_config = vz.ProblemStatement(
        search_space=test_studies.flat_space_with_all_types(),
        metric_information=[
            vz.MetricInformation('x1', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
        ],
    )

    self._designer = random_designer_lib.RandomDesigner(
        self._study_config.search_space, seed=0
    )
    self._trials = test_runners.run_with_random_metrics(
        self._designer, self._study_config, iters=1, batch_size=10
    )
    self.maxDiff = None

  def test_back_to_back_conversion(self):
    converter = jnpc.TrialToContinuousAndCategoricalConverter.from_study_config(
        self._study_config
    )
    self.assertSequenceEqual(
        [t.parameters for t in self._trials],
        converter.to_parameters(converter.to_features(self._trials)),
    )

  def test_dtype(self):
    space = vz.SearchSpace()
    root = space.root
    root.add_float_param('double', -2.0, 2.0)
    root.add_int_param('integer', -2, 2)
    root.add_categorical_param('categorical', ['b', 'c'])
    converter = jnpc.TrialToContinuousAndCategoricalConverter.from_study_config(
        vz.ProblemStatement(search_space=space)
    )
    self.assertEqual(converter.dtype.continuous, np.float64)
    self.assertEqual(converter.dtype.categorical, np.int32)

  def test_parameter_continuify(self):
    space = vz.SearchSpace()
    root = space.root
    root.add_float_param('double', -2.0, 2.0)
    root.add_int_param('integer', -2, 2)
    root.add_categorical_param('categorical', ['b', 'c'])
    root.add_discrete_param('discrete', [-1.0, 2.0, 3.0])

    converter = jnpc.TrialToContinuousAndCategoricalConverter.from_study_config(
        vz.ProblemStatement(search_space=space)
    )
    trial = vz.Trial(
        parameters={
            'double': vz.ParameterValue(3.0),
            'integer': vz.ParameterValue(-1),
            'discrete': vz.ParameterValue(2.0),
            'categorical': vz.ParameterValue('d'),
        }
    )
    expected = types.ContinuousAndCategoricalArray(
        continuous=np.array([[1.25, 0.25, 0.75]]), categorical=np.array([[2]])
    )
    actual = converter.to_features([trial])
    np.testing.assert_equal(actual.continuous, expected.continuous)
    np.testing.assert_equal(actual.categorical, expected.categorical)

  def test_multi_metrics(self):
    search_space = vz.SearchSpace()
    search_space.root.add_float_param('x', 0.0, 1.0)
    problem = vz.ProblemStatement(
        search_space=search_space,
        metric_information=vz.MetricsConfig(
            metrics=[
                vz.MetricInformation(
                    'obj1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
                vz.MetricInformation(
                    'obj2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
                vz.MetricInformation(
                    'obj3', goal=vz.ObjectiveMetricGoal.MINIMIZE
                ),
            ]
        ),
    )
    trial1 = vz.Trial()
    trial2 = vz.Trial()
    trial1.final_measurement = vz.Measurement(
        metrics={'obj1': 1.0, 'obj2': 2.0, 'obj3': 3.0}
    )
    trial2.final_measurement = vz.Measurement(
        metrics={'obj1': -1.0, 'obj2': 5.0, 'obj3': 0.0}
    )
    converter = jnpc.TrialToContinuousAndCategoricalConverter.from_study_config(
        problem
    )
    # Notice that the sign is flipped for MINIMIZE objective.
    expected_labels = np.array([[1.0, 2.0, -3.0], [-1.0, 5.0, 0.0]])
    np.testing.assert_equal(
        converter.to_labels([trial1, trial2]), expected_labels
    )
    self.assertEqual(converter.metric_specs[0].name, 'obj1')
    self.assertEqual(converter.metric_specs[1].name, 'obj2')
    self.assertEqual(converter.metric_specs[2].name, 'obj3')


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for atari100k."""
from absl import logging

from vizier import pyvizier
from vizier._src.algorithms.designers import random

from absl.testing import absltest
from absl.testing import parameterized


class Atari100KTest(parameterized.TestCase):

  @absltest.skip("ALE ROMS must be installed manually.")
  @parameterized.parameters('DER', 'DrQ', 'DrQ_eps', 'OTRainbow')
  @absltest.skip('Jax versioning not updated in Dopamine.')
  def test_e2e_evaluation(self, agent_name):
    from vizier._src.benchmarks.experimenters import atari100k_experimenter  # pylint: disable=g-import-not-at-top

    initial_gin_bindings = {
        'Runner.training_steps': 2,
        'MaxEpisodeEvalRunner.num_eval_episodes': 2,
        'Runner.num_iterations': 2,
        'Runner.max_steps_per_episode': 2,
        'JaxDQNAgent.min_replay_history': 2,
        'OutOfGraphPrioritizedReplayBuffer.replay_capacity': 1000,
    }
    experimenter = atari100k_experimenter.Atari100kExperimenter(
        game_name='Pong',
        agent_name=agent_name,
        initial_gin_bindings=initial_gin_bindings,
    )

    designer = random.RandomDesigner(
        experimenter.problem_statement().search_space, seed=None
    )

    suggestions = designer.suggest(2)
    trials = [suggestion.to_trial() for suggestion in suggestions]
    experimenter.evaluate(trials)
    # Trials should be completed in place.
    for trial in trials:
      self.assertEqual(trial.status, pyvizier.TrialStatus.COMPLETED)
      logging.info('Evaluated Trial: %s', trial)
      if trial.final_measurement:
        value = trial.final_measurement.metrics['eval_average_return'].value
        self.assertGreaterEqual(value, 0.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier._src.benchmarks.experimenters.synthetic import multiobjective_optproblems as mopt
from vizier._src.benchmarks.testing import experimenter_testing
from absl.testing import absltest
from absl.testing import parameterized


class WFGxperimenterTest(parameterized.TestCase):

  @parameterized.parameters(
      ("WFG1",),
      ("WFG2",),
      ("WFG3",),
      ("WFG4",),
      ("WFG5",),
      ("WFG6",),
      ("WFG7",),
      ("WFG8",),
      ("WFG9",),
  )
  def test_e2e_smoke(self, name):
    exptr_factory = mopt.WFGExperimenterFactory(name, dim=6, num_objectives=3)
    experimenter_testing.assert_evaluates_random_suggestions(
        self, exptr_factory()
    )


class DTLZExperimenterTest(parameterized.TestCase):

  @parameterized.parameters(
      ("DTLZ1",),
      ("DTLZ2",),
      ("DTLZ3",),
      ("DTLZ4",),
      ("DTLZ5",),
      ("DTLZ6",),
      ("DTLZ7",),
  )
  def test_e2e_smoke(self, name):
    exptr_factory = mopt.DTLZExperimenterFactory(name, dim=4, num_objectives=3)
    experimenter_testing.assert_evaluates_random_suggestions(
        self, exptr_factory()
    )


class ZDTExperimenterTest(parameterized.TestCase):

  @parameterized.parameters(
      ("ZDT1",),
      ("ZDT2",),
      ("ZDT3",),
      ("ZDT4",),
      ("ZDT6",),
  )
  def test_e2e_smoke(self, name):
    exptr_factory = mopt.ZDTExperimenterFactory(name, dim=5)
    experimenter_testing.assert_evaluates_random_suggestions(
        self, exptr_factory()
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Bayesian Optimization of Combinatorial Structures (BOCS) from https://arxiv.org/abs/1806.08838.

Code is a cleaned-up exact version from /BOCSpy/ in
https://github.com/baptistar/BOCS.
"""
# pylint:disable=invalid-name
import abc
import itertools
from typing import Callable, Optional, Sequence, Tuple, Union

from absl import logging
import cvxpy as cvx
import numpy as np

from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random

FloatType = Union[float, np.float32, np.float64]


class _BayesianHorseshoeLinearRegression:
  """Computes conditional poster parameter distributions from a sparsity-inducing prior."""

  def _fastmvg(self, Phi: np.ndarray, alpha: np.ndarray,
               D: np.ndarray) -> np.ndarray:
    """Fast sampler for multivariate Gaussian distributions (large p, p > n) of the form N(mu, S).

    We have:
      mu = S Phi' y
      S  = inv(Phi'Phi + inv(D))
    Reference: https://arxiv.org/abs/1506.04778

    Args:
      Phi:
      alpha:
      D:

    Returns:
    """
    n, p = Phi.shape
    d = np.diag(D)
    u = np.random.randn(p) * np.sqrt(d)
    delta = np.random.randn(n)
    v = np.dot(Phi, u) + delta
    mult_vector = np.vectorize(np.multiply)
    Dpt = mult_vector(Phi.T, d[:, np.newaxis])
    w = np.linalg.solve(np.matmul(Phi, Dpt) + np.eye(n), alpha - v)
    return u + np.dot(Dpt, w)

  def _fastmvg_rue(self, Phi: np.ndarray, PtP: np.ndarray, alpha: np.ndarray,
                   D: np.ndarray) -> np.ndarray:
    """Another sampler for multivariate Gaussians (small p) of the form N(mu, S).

    We have:
      mu = S Phi' y
      S  = inv(Phi'Phi + inv(D))
    Here, PtP = Phi'*Phi (X'X is precomputed).

    Reference: https://www.jstor.org/stable/2680602

    Args:
      Phi:
      PtP:
      alpha:
      D:

    Returns:
    """
    p = Phi.shape[1]
    Dinv = np.diag(1. / np.diag(D))

    # Regularize PtP + Dinv matrix for small negative eigenvalues.
    try:
      L = np.linalg.cholesky(PtP + Dinv)
    except np.linalg.LinAlgError:
      mat = PtP + Dinv
      Smat = (mat + mat.T) / 2.
      maxEig_Smat = np.max(np.linalg.eigvals(Smat))
      L = np.linalg.cholesky(Smat + maxEig_Smat * 1e-15 * np.eye(Smat.shape[0]))

    v = np.linalg.solve(L, np.dot(Phi.T, alpha))
    m = np.linalg.solve(L.T, v)
    w = np.linalg.solve(L.T, np.random.randn(p))
    return m + w

  def regress(
      self,
      X: np.ndarray,
      Y: np.ndarray,
      nsamples: int = 1000,
      burnin: int = 0,
      thin: int = 1
  ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Implementation of the Bayesian horseshoe linear regression hierarchy.

    References:
      https://arxiv.org/abs/1508.03884
      https://www.jstor.org/stable/25734098
      (c) Copyright Enes Makalic and Daniel F. Schmidt, 2015
      Adapted to python by Ricardo Baptista, 2018

    Args:
      X: regressor matrix [n x p]
      Y: response vector  [n x 1]
      nsamples: number of samples for the Gibbs sampler (nsamples > 0)
      burnin: number of burnin (burnin >= 0)
      thin: thinning (thin >= 1)

    Returns:
      beta: regression parameters  [p x nsamples]
      b0: regression param. for constant [1 x nsamples]
      s2: noise variance sigma^2 [1 x nsamples]
      t2: hypervariance tau^2    [1 x nsamples]
      l2: hypervariance lambda^2 [p x nsamples]
    """

    n, p = X.shape

    # Standardize y's
    muY = np.mean(Y)
    Y = Y - muY

    # Return values
    beta = np.zeros((p, nsamples))
    s2 = np.zeros((1, nsamples))
    t2 = np.zeros((1, nsamples))
    l2 = np.zeros((p, nsamples))

    # Initial values
    sigma2 = 1.
    lambda2 = np.random.uniform(size=p)
    tau2 = 1.
    nu = np.ones(p)
    xi = 1.

    # pre-compute X'*X (used with fastmvg_rue)
    XtX = np.matmul(X.T, X)

    # Gibbs sampler
    k = 0
    iter_count = 0
    while k < nsamples:

      # Sample from the conditional posterior distribution
      sigma = np.sqrt(sigma2)
      lambda_star = tau2 * np.diag(lambda2)
      # Determine best sampler for conditional posterior of beta's
      if (p > n) and (p > 200):
        b = self._fastmvg(X / sigma, Y / sigma, sigma2 * lambda_star)
      else:
        b = self._fastmvg_rue(X / sigma, XtX / sigma2, Y / sigma,
                              sigma2 * lambda_star)

      # Sample sigma2
      e = Y - np.dot(X, b)
      shape = (n + p) / 2.
      scale = np.dot(e.T, e) / 2. + np.sum(b**2 / lambda2) / tau2 / 2.
      sigma2 = 1. / np.random.gamma(shape, 1. / scale)

      # Sample lambda2
      scale = 1. / nu + b**2. / 2. / tau2 / sigma2
      lambda2 = 1. / np.random.exponential(1. / scale)

      # Sample tau2
      shape = (p + 1.) / 2.
      scale = 1. / xi + np.sum(b**2. / lambda2) / 2. / sigma2
      tau2 = 1. / np.random.gamma(shape, 1. / scale)

      # Sample nu
      scale = 1. + 1. / lambda2
      nu = 1. / np.random.exponential(1. / scale)

      # Sample xi
      scale = 1. + 1. / tau2
      xi = 1. / np.random.exponential(1. / scale)

      # Store samples
      iter_count += 1
      if iter_count > burnin:
        # thinning
        if (iter_count % thin) == 0:
          beta[:, k] = b
          s2[:, k] = sigma2
          t2[:, k] = tau2
          l2[:, k] = lambda2
          k = k + 1

    b0 = muY
    return beta, b0, s2, t2, l2


class _GibbsLinearRegressor:
  """Stateful Gibbs sampler."""

  def __init__(self, order: int, num_gibbs_retries: int = 10):
    self._order = order
    self._num_gibbs_retries = num_gibbs_retries

    # Below are stateful attributes calulated after `regress()`.
    self._alpha: Optional[np.ndarray] = None
    self._num_vars: Optional[int] = None
    self._X_inf: Optional[np.ndarray] = None
    self._y_inf: Optional[np.ndarray] = None

  def _preprocess(
      self,
      X: np.ndarray,
      Y: np.ndarray,
      inf_threshold: float = 1e6
  ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Preprocess data to ensure unique points and remove outliers."""
    # Limit data to unique points.
    unique_X, X_idx = np.unique(X, axis=0, return_index=True)
    unique_Y = Y[X_idx]

    # separate samples based on Inf output
    Y_Infidx = np.where(np.abs(unique_Y) > inf_threshold)[0]
    Y_nInfidx = np.setdiff1d(np.arange(len(unique_Y)), Y_Infidx)

    X_train = unique_X[Y_nInfidx, :]
    Y_train = unique_Y[Y_nInfidx]

    # Large-value outliers.
    X_inf = X[Y_Infidx, :]
    Y_inf = Y[Y_Infidx]

    return X_train, Y_train, X_inf, Y_inf

  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    """Computes and saves alpha (regressor coefficients from the data."""
    # Preprocess data for training and store relevant data.
    X_train, Y_train, self._X_inf, self._Y_inf = self._preprocess(X, Y)
    self._num_vars = X_train.shape[1]

    # Create matrix with all covariates based on order.
    X_train = self._order_effects(X_train)
    (nSamps, nCoeffs) = X_train.shape

    # Check if X_train contains columns w/ zeros and find corresponding indices.
    check_zero = np.all(X_train == np.zeros((nSamps, 1)), axis=0)
    idx_nnzero = np.where(check_zero == False)[0]  # pylint:disable=singleton-comparison,g-explicit-bool-comparison

    # Remove columns of zeros in X_train.
    if np.any(check_zero):
      X_train = X_train[:, idx_nnzero]

    # Run Gibbs sampler.
    bhs = _BayesianHorseshoeLinearRegression()

    counter = 0
    while counter < self._num_gibbs_retries:
      # re-run if there is an error during sampling
      counter += 1
      try:
        alphaGibbs, a0, _, _, _ = bhs.regress(X_train, Y_train)
        # run until alpha matrix does not contain any NaNs
        if not np.isnan(alphaGibbs).any():
          break
      except np.linalg.LinAlgError:
        logging.error('Numerical error during Gibbs sampling. Trying again.')
        continue

    if counter >= self._num_gibbs_retries:
      raise ValueError('Gibbs sampling failed for %d tries.' %
                       self._num_gibbs_retries)

    # append zeros back - note alpha(1,:) is linear intercept
    alpha_pad = np.zeros(nCoeffs)
    alpha_pad[idx_nnzero] = alphaGibbs[:, -1]
    self._alpha = np.append(a0, alpha_pad)

  @property
  def alpha(self) -> np.ndarray:
    if self._alpha is None:
      raise ValueError('You first need to call `regress()` on the data.')
    return self._alpha

  @property
  def num_vars(self) -> int:
    if self._num_vars is None:
      raise ValueError('You first need to call `regress()` on the data.')
    return self._num_vars

  def surrogate_model(self, x: np.ndarray) -> FloatType:
    """Surrogate model.

    Args:
      x: Should only contain one row.

    Returns:
      Surrogate objective.
    """

    # Generate x_all (all basis vectors) based on model order.
    x_all = np.append(1, self._order_effects(x))

    # check if previous x led to Inf output (if so, barrier=Inf)
    barrier = 0.
    if self._X_inf.shape[0] != 0 and np.equal(x, self._X_inf).all(axis=1).any():
      barrier = np.inf

    return np.dot(x_all, self._alpha) + barrier

  def _order_effects(self, X: np.ndarray) -> np.ndarray:
    """Function computes data matrix for all coupling."""

    # Find number of variables
    n_samp, n_vars = X.shape

    # Generate matrix to store results
    x_allpairs = X

    for ord_i in range(2, self._order + 1):

      # generate all combinations of indices (without diagonals)
      offdProd = np.array(
          list(itertools.combinations(np.arange(n_vars), ord_i)))

      # generate products of input variables
      x_comb = np.zeros((n_samp, offdProd.shape[0], ord_i))
      for j in range(ord_i):
        x_comb[:, :, j] = X[:, offdProd[:, j]]
      x_allpairs = np.append(x_allpairs, np.prod(x_comb, axis=2), axis=1)

    return x_allpairs


class AcquisitionOptimizer(abc.ABC):
  """Base class for BOCS acquisition optimizers."""

  def __init__(self, lin_reg: _GibbsLinearRegressor, lamda: float = 1e-4):
    self._lin_reg = lin_reg
    self._num_vars = self._lin_reg.num_vars
    self._lamda = lamda

  @abc.abstractmethod
  def argmin(self) -> np.ndarray:
    """Computes argmin using the regressor."""
    pass


class SimulatedAnnealing(AcquisitionOptimizer):
  """Simulated Annealing solver."""

  def __init__(self,
               lin_reg: _GibbsLinearRegressor,
               lamda: float = 1e-4,
               num_iters: int = 10,
               num_reruns: int = 5,
               initial_temp: float = 1.0,
               annealing_factor: float = 0.8):
    super().__init__(lin_reg=lin_reg, lamda=lamda)
    self._num_iters = num_iters
    self._num_reruns = num_reruns
    self._initial_temp = initial_temp
    self._annealing_factor = annealing_factor

  def argmin(self) -> np.ndarray:
    """Computes argmin via multiple rounds of Simulated Annealing."""
    SA_model = np.zeros((self._num_reruns, self._num_vars))
    SA_obj = np.zeros(self._num_reruns)

    penalty = lambda x: self._lamda * np.sum(x, axis=1)
    acquisition_fn = lambda x: self._lin_reg.surrogate_model(x) + penalty(x)

    for j in range(self._num_reruns):
      optModel, objVals = self._optimization_loop(acquisition_fn)
      SA_model[j, :] = optModel[-1, :]
      SA_obj[j] = objVals[-1]

    # Find optimal solution
    min_idx = np.argmin(SA_obj)
    x_new = SA_model[min_idx, :]
    return x_new

  def _optimization_loop(
      self, objective: Callable[[np.ndarray], FloatType]
  ) -> Tuple[np.ndarray, np.ndarray]:
    """Single optimization round of Simulated Annealing."""

    # Declare vectors to save solutions
    model_iter = np.zeros((self._num_iters, self._num_vars))
    obj_iter = np.zeros(self._num_iters)

    # Set initial temperature and cooling schedule
    T = self._initial_temp
    cool = lambda t: self._annealing_factor * t

    # Set initial condition and evaluate objective
    old_x = np.zeros((1, self._num_vars))
    old_obj = objective(old_x)

    # Set best_x and best_obj
    best_x = old_x
    best_obj = old_obj

    # Run simulated annealing
    for t in range(self._num_iters):

      # Decrease T according to cooling schedule.
      T = cool(T)

      # Find new sample
      flip_bit = np.random.randint(self._num_vars)
      new_x = old_x.copy()
      new_x[0, flip_bit] = 1. - new_x[0, flip_bit]

      # Evaluate objective function.
      new_obj = objective(new_x)

      # Update current solution iterate.
      if (new_obj < old_obj) or (np.random.rand() < np.exp(
          (old_obj - new_obj) / T)):
        old_x = new_x
        old_obj = new_obj

      # Update best solution
      if new_obj < best_obj:
        best_x = new_x
        best_obj = new_obj

      # Save solution
      model_iter[t, :] = best_x
      obj_iter[t] = best_obj

    return model_iter, obj_iter


class SemiDefiniteProgramming(AcquisitionOptimizer):
  """SDP solver for quadratic acquisition functions."""

  def __init__(self,
               lin_reg: _GibbsLinearRegressor,
               lamda: float = 1e-4,
               num_repeats: int = 100):
    super().__init__(lin_reg=lin_reg, lamda=lamda)
    self._num_repeats = num_repeats

  def argmin(self) -> np.ndarray:
    """Perform SDP over the quadratic xt*A*x + bt*x.

    (A,b) is recovered from alpha.

    Returns:
      Argmin of the SDP problem.
    """
    alpha = self._lin_reg.alpha

    # Extract vector of coefficients
    b = alpha[1:self._num_vars + 1] + self._lamda
    a = alpha[self._num_vars + 1:]

    # Get indices for quadratic terms.
    idx_prod = np.array(
        list(itertools.combinations(np.arange(self._num_vars), 2)))
    n_idx = idx_prod.shape[0]

    # Check number of coefficients
    if a.size != n_idx:
      raise ValueError('Number of Coefficients does not match indices!')

    # Convert a to matrix form
    A = np.zeros((self._num_vars, self._num_vars))
    for i in range(n_idx):
      A[idx_prod[i, 0], idx_prod[i, 1]] = a[i] / 2.
      A[idx_prod[i, 1], idx_prod[i, 0]] = a[i] / 2.

    # Convert to standard form.
    bt = b / 2. + np.dot(A, np.ones(self._num_vars)) / 2.
    bt = bt.reshape((self._num_vars, 1))
    At = np.vstack((np.append(A / 4., bt / 2., axis=1), np.append(bt.T, 2.)))

    # Run SDP relaxation.
    X = cvx.Variable((self._num_vars + 1, self._num_vars + 1), PSD=True)
    obj = cvx.Minimize(cvx.trace(cvx.matmul(At, X)))
    constraints = [cvx.diag(X) == np.ones(self._num_vars + 1)]
    prob = cvx.Problem(obj, constraints)
    prob.solve(solver=cvx.CVXOPT)

    # Extract vectors and compute Cholesky.
    try:
      L = np.linalg.cholesky(X.value)
    except np.linalg.LinAlgError:
      XpI = X.value + 1e-15 * np.eye(self._num_vars + 1)
      L = np.linalg.cholesky(XpI)

    suggest_vect = np.zeros((self._num_vars, self._num_repeats))
    obj_vect = np.zeros(self._num_repeats)

    for kk in range(self._num_repeats):
      # Generate a random cutting plane vector (uniformly distributed on the
      # unit sphere - normalized vector)
      r = np.random.randn(self._num_vars + 1)
      r = r / np.linalg.norm(r)
      y_soln = np.sign(np.dot(L.T, r))

      # Convert solution to original domain and assign to output vector.
      suggest_vect[:, kk] = (y_soln[:self._num_vars] + 1.) / 2.
      obj_vect[kk] = np.dot(
          np.dot(suggest_vect[:, kk].T, A), suggest_vect[:, kk]) + np.dot(
              b, suggest_vect[:, kk])

    # Find optimal rounded solution.
    opt_idx = np.argmin(obj_vect)
    return suggest_vect[:, opt_idx]


AcqusitionOptimizerFactory = Callable[[_GibbsLinearRegressor, float],
                                      AcquisitionOptimizer]


class BOCSDesigner(vza.Designer):
  """BOCS Designer."""

  def __init__(self,
               problem_statement: vz.ProblemStatement,
               order: int = 2,
               acquisition_optimizer_factory:
               AcqusitionOptimizerFactory = SemiDefiniteProgramming,
               lamda: float = 1e-4,
               num_initial_randoms: int = 10):
    """Init.

    Args:
      problem_statement: Must use a boolean search space.
      order: Statistical model order.
      acquisition_optimizer_factory: Which acquisition optimizer to use.
      lamda: Sparsity regularization coefficient.
      num_initial_randoms: Number of initial random suggestions for seeding the
        model.
    """

    if problem_statement.search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.')
    for p_config in problem_statement.search_space.parameters:
      if p_config.external_type != vz.ExternalType.BOOLEAN:
        raise ValueError('Only boolean search spaces are supported.')

    self._problem_statement = problem_statement
    self._metric_name = self._problem_statement.metric_information.item().name
    self._search_space = problem_statement.search_space
    self._current_index = 0

    self._order = order
    self._acquisition_optimizer_factory = acquisition_optimizer_factory
    self._lamda = lamda
    self._num_initial_randoms = num_initial_randoms

    self._trials = []

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    self._trials += tuple(completed.trials)

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Core BOCS method.

    Initially will use random search for the first `num_initial_randoms`
    suggestions, and then will use SDP or Simulated Annealing for acqusition
    minimization.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1
    if count > 1:
      raise ValueError('This designer does not support batched suggestions.')

    if len(self._trials) < self._num_initial_randoms:
      random_designer = random.RandomDesigner(self._search_space)
      return random_designer.suggest(count)

    X = []
    Y = []
    for t in self._trials:
      single_x = [
          float(t.parameters.get_value(p.name) == 'True')
          for p in self._search_space.parameters
      ]
      single_y = t.final_measurement.metrics[self._metric_name].value
      X.append(single_x)
      Y.append(single_y)
    X = np.array(X)
    Y = np.array(Y)

    if self._problem_statement.metric_information.item(
    ).goal == vz.ObjectiveMetricGoal.MAXIMIZE:
      Y = -Y

    # Train initial statistical model
    lin_reg = _GibbsLinearRegressor(self._order)
    lin_reg.regress(X, Y)

    # Run acquisition optimization.
    optimizer = self._acquisition_optimizer_factory(lin_reg, self._lamda)
    x_new = optimizer.argmin()

    parameters = vz.ParameterDict()
    for i, p in enumerate(self._search_space.parameters):
      parameters[p.name] = 'True' if x_new[i] == 1.0 else 'False'
    return [vz.TrialSuggestion(parameters=parameters)]



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""End-to-end tests for Vizier backend for PyGlove."""

import pyglove as pg
from vizier._src.pyglove import oss_vizier
from vizier._src.service import clients as pyvizier_clients
from absl.testing import absltest


pyvizier_clients.environment_variables.servicer_use_sql_ram()


class PygloveTest(absltest.TestCase):
  """Tests for using Vizier as PyGlove backend."""

  def test_sample(self):
    oss_vizier.init('my_study')

    examples = []
    for x, f in pg.sample(
        pg.oneof([1, 2, 3]), pg.geno.Random(seed=1), num_examples=3
    ):
      f(x)
      examples.append(x)

    self.assertEqual(examples, [1, 3, 1])
    result = pg.poll_result('')
    self.assertLen(result.trials, 3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utilities for computing exploration scores."""

from typing import Iterable, Optional

import numpy as np
import scipy
from vizier import pyvizier as vz

BenchmarkResults = dict[str, dict[str, dict[int, vz.ProblemAndTrials]]]


def compute_parameter_entropy(
    parameter_config: vz.ParameterConfig,
    parameter_values: Iterable[Optional[vz.ParameterValue]],
) -> float:
  """Computes the entropy of parameter values.

  Args:
    parameter_config: The parameter config.
    parameter_values: Values of a parameter.

  WARNING: Entropy estimation accuracy depends on the sample size, so to compare
  the entropies of two `parameter_values`, make sure they have the same size.

  Returns:
    The entropy of parameter values.
  """
  values = [pv.value for pv in parameter_values if pv is not None]
  if not values:
    return 0.0
  if parameter_config.type in [
      vz.ParameterType.CATEGORICAL,
      vz.ParameterType.DISCRETE,
  ] and hasattr(parameter_config, 'feasible_values'):
    if any([value not in parameter_config.feasible_values for value in values]):
      raise ValueError(
          f'Parameter values: {parameter_values} contain out-of-bound values.'
          f' Feasible values: {parameter_config.feasible_values}'
      )
    _, counts = np.unique(values, return_counts=True)
  elif hasattr(parameter_config, 'bounds'):
    min_val = parameter_config.bounds[0]
    max_val = parameter_config.bounds[1]
    if any([value < min_val or value > max_val for value in values]):
      raise ValueError(
          f'Parameter values: {parameter_values} contain out-of-bound values.'
          f' Bound: [{min_val}, {max_val}]'
      )
    if parameter_config.type == vz.ParameterType.INTEGER:
      _, counts = np.unique(values, return_counts=True)
    else:
      # Sets the number of fixed-width bins as c * sample_size ** (1.0 / 3.0).
      # The cubic-root dependency on the sample size appears in several common
      # bin-size selection strategies, e.g.
      # https://en.wikipedia.org/wiki/Histogram#Number_of_bins_and_width
      # The multiplier `c` is chosen such that for a small sample size, say 100,
      # we still get a reasonable number of bins, say 30.
      alpha = 1.0 / 3.0
      c = 30.0 / (100**alpha)
      n_sample = len(values)
      # We also ensure that the number of bins is at most the sample size.
      num_bins = np.min((int(c * n_sample**alpha), n_sample))
      counts, _ = np.histogram(
          values,
          bins=np.linspace(
              min_val,
              max_val,
              num=num_bins + 1,
              dtype=np.float32,
          ),
      )
  else:
    raise ValueError(
        'Invalid parameter config: either `feasible_values` or'
        '`bounds` is expected to be set, but both are unset. '
        f'Parameter config: {parameter_config}'
    )
  return float(scipy.stats.entropy(counts))


def compute_average_marginal_parameter_entropy(
    results: BenchmarkResults,
) -> float:
  """Computes the average marginal parameter entropy across results.

  Computes the marginal entropy of every parameter in every study, and then
  returns the average marginal entropy over all parameters and all studies.

  Args:
    results: Benchmark results.

  Returns:
    Average marginal parameter entropy.
  """
  marginal_param_entropies = []
  for _, spec_gen_results in results.items():
    for _, spec_results in spec_gen_results.items():
      for _, study in spec_results.items():
        for param_config in study.problem.search_space.parameters:
          param_values = [
              trial.parameters.get(param_config.name) for trial in study.trials
          ]
          marginal_param_entropies.append(
              compute_parameter_entropy(
                  parameter_config=param_config, parameter_values=param_values
              )
          )

  return np.mean(marginal_param_entropies)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Multimetric utility classes for Studies/Trials."""

from typing import Iterable, Sequence
import numpy as np
from vizier import pyvizier


class SafetyChecker:
  """Returns the safety status of Measurements/Trials."""

  def __init__(self, metrics_config: pyvizier.MetricsConfig):
    """Initializes with metric information (can have non-safety info).

    Example usage from a study:

      pyconfig = pyvizier.StudyConfig.from_proto(study.study_config)
      checker = multimetric_util.SafetyChecker(pyconfig.metric_information)
      are_trials_safe = checker.are_trials_safe(
              converters.TrialConverter.from_protos(study.trials))
    Args:
      metrics_config: MetricsConfig for a Study.
    """
    self._objective_metrics = list(
        metrics_config.of_type(pyvizier.MetricType.OBJECTIVE)
    )
    self._safety_metrics = list(
        metrics_config.of_type(pyvizier.MetricType.SAFETY))

  def are_trials_safe(self, trials: Iterable[pyvizier.Trial]) -> Iterable[bool]:
    """Returns whether the final measurements of Trials are safe.

    Args:
      trials: Iterable of pyvizier Trials.

    Returns:
      Iterable of whether Trials are safe. Safety is assumed if no final
      measurement is found or safety metric is not found in final measurement.
    """
    return self.are_measurements_safe([
        t.final_measurement if t.final_measurement else pyvizier.Measurement()
        for t in trials
    ])

  def are_measurements_safe(
      self, measurements: Iterable[pyvizier.Measurement]) -> Iterable[bool]:
    """Returns whether measurements are safe.

    Args:
      measurements: Iterable of pyvizier measurements.

    Returns:
      Iterable of whether measurements are safe. Safety is assumed if safety
      metric is not found in measurement.
    """
    safety_results = []
    for measurement in measurements:
      is_safe = True
      for safety_metric in self._safety_metrics:
        # If safety metric is not in metrics, we assume it is safe.
        if safety_metric.name not in measurement.metrics:
          continue

        # Otherwise, we check if there is a safety violation.
        metric_value = measurement.metrics[safety_metric.name].value
        if safety_metric.goal == pyvizier.ObjectiveMetricGoal.MAXIMIZE:
          is_safe = is_safe and metric_value >= safety_metric.safety_threshold
        else:
          is_safe = is_safe and metric_value <= safety_metric.safety_threshold
      safety_results.append(is_safe)
    return safety_results

  def warp_unsafe_trials(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Sequence[pyvizier.Trial]:
    """Sets the final measurements of unsafe Trials to worst values in place.

    Args:
      trials: Sequence of pyvizier Trials.

    Returns:
      Trials whose final measurements are warped to the worst value (inf/-inf)
      for all OBJECTIVE metrics in unsafe Trials.
    """
    are_safe = self.are_trials_safe(trials)
    objective_goals = {m.name: m.goal for m in self._objective_metrics}
    for is_safe, trial in zip(are_safe, trials):
      if is_safe:
        continue
      if trial.final_measurement is None:
        continue

      for name in trial.final_measurement.metrics.keys():
        if name in objective_goals:
          worst_value = (
              np.inf
              if objective_goals[name] == pyvizier.ObjectiveMetricGoal.MINIMIZE
              else -np.inf
          )
          trial.final_measurement.metrics[name] = pyvizier.Metric(
              value=worst_value
          )

    return trials
